/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2010-2014 Intel Corporation
 */

#ifndef __INCLUDE_RTE_LRU_X86_H__
#define __INCLUDE_RTE_LRU_X86_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#include <rte_config.h>
#include <rte_vect.h>

#ifndef RTE_TABLE_HASH_LRU_STRATEGY
#define RTE_TABLE_HASH_LRU_STRATEGY                        3
#endif

#if RTE_TABLE_HASH_LRU_STRATEGY == 3

#define lru_init(bucket)						\
	{ bucket->lru_list = ~0LLU; }

static inline int
f_lru_pos(uint64_t lru_list)
{
	xmm_t lst = (xmm_t) {.i64 = { lru_list, (uint64_t)-1} };
	uint16_t min_val, i, j = 0;

	min_val = lst.u16[0];
	for (i = 0; i < 8; i++) {
		if (min_val > lst.u16[i]) {
			min_val = lst.u16[i];
			j = i;
		}
	}

	return j;
}
#define lru_pos(bucket) f_lru_pos(bucket->lru_list)

#define lru_update(bucket, mru_val)					\
do {									\
	const uint64_t orvals[] = {0xFFFFLLU, 0xFFFFLLU << 16,		\
		0xFFFFLLU << 32, 0xFFFFLLU << 48, 0LLU};		\
	const uint64_t decs[] = {0x1000100010001LLU, 0};		\
	xmm_t v_zeros = (xmm_t) {.u64 = {0, 0} };			\
	xmm_t lru = v_zeros, vdec = v_zeros;				\
	lru.v2i64 = __msa_insert_d(lru.v2i64, 0, bucket->lru_list);	\
	vdec.v2i64 = __msa_insert_d(vdec.v2i64, 0, decs[mru_val>>2]);	\
	lru.v8u16 = __msa_subs_u_h(lru.v8u16, vdec.v8u16);		\
	bucket->lru_list = __msa_copy_u_d(lru.v2i64, 0) | orvals[mru_val];	\
} while (0)

#endif

#ifdef __cplusplus
}
#endif

#endif
